import asyncio
import itertools
import json

from collections import OrderedDict
from typing import Any, Iterator, Optional

from glebby.board import BoardGenerator

class Client:
    outgoing_queue: asyncio.Queue[str]
    id: int
    name: str

    def __init__(self, client_id: int, outgoing_queue: asyncio.Queue[str]):
        self.outgoing_queue = outgoing_queue
        self.id = client_id
        self.name = ''
    
    async def put_outgoing_message(self, message: str) -> None:
        await self.outgoing_queue.put(message)
    
    def get_dto(self) -> dict[str, Any]:
        return { 'id': self.id, 'name': self.id }

class Model:
    clients: OrderedDict[int, Client]
    id_generator: Iterator[int]
    board_generator: BoardGenerator
    board: Optional[list[list[str]]]

    def __init__(self) -> None:
        self.clients = OrderedDict()
        self.id_generator = itertools.count(start=0, step=1)
        self.board_generator = BoardGenerator(None)
        self.board = None
    
    async def add_client(self, outgoing_queue: asyncio.Queue[str]) -> int:
        client_id = next(self.id_generator)
        await self.broadcast(client_id, { 'type': 'join' })

        self.clients[client_id] = Client(client_id, outgoing_queue)
        print(f'<{client_id}|join>')

        await self.send_to(None, client_id, {
            'type': 'init',
            'state': self.get_state_dto(client_id)
        })

        return client_id
    
    async def remove_client(self, client_id: int) -> None:
        del self.clients[client_id]
        await self.broadcast(client_id, { 'type': 'leave' })
        
        print(f'<{client_id}|leave>')
    
    async def send_to(self, client_id_from: Optional[int], client_id_to: int, payload: Any) -> None:
        await self.clients[client_id_to].put_outgoing_message(json.dumps({
            'from': client_id_from,
            'payload': payload
        }))
    
    async def broadcast(self, client_id_from: Optional[int], payload: Any) -> None:
        for client_id_to in self.clients:
            await self.send_to(client_id_from, client_id_to, payload)
    
    async def handle_incoming(self, client_id: int, message: str) -> None:
        print(f'<{client_id}|data> {message}')
        payload = json.loads(message)
        
        if payload['type'] == 'set-name':
            self.clients[client_id].name = payload['name']
            await self.broadcast(client_id, {
                'type': 'set-name',
                'name': payload['name']
            })
        elif payload['type'] == 'chat':
            await self.broadcast(client_id, {
                'type': 'chat',
                'message': payload['message']
            })
        elif payload['type'] == 'roll':
            self.board = self.board_generator.generate_board()
            await self.broadcast(client_id, {
                'type': 'roll',
                'board': self.board
            })
        else:
            print("Unhandled!")
    
    def get_state_dto(self, client_id: int) -> Any:
        return {
            'yourId': client_id,
            'players': [
                client.get_dto()
                for client in self.clients.values()
            ],
            'board': self.board
        }

